#include "bonus_points.hpp"

#include "util/exception.hpp"
#include "util/heuristics.hpp"
#include "util/strings.hpp"

#include <algorithm>
#include <optional>
#include <ranges>
#include <stdexcept>
#include <string>

BonusPoints::BonusPoints(const CsvDump &c) noexcept(false) : mBonusPoints(c) {
    if (c.data.empty())
        throw BonusParseException(str::error::empty);

    if (c.data[0].empty())
        throw BonusParseException(str::error::noHeader);

    // Quick check: reading wrong file type?
    {
        const auto fileType =
            getApproximateFileType(QString::fromStdString(c.data[0][0]));

        switch (fileType) {
        case FileType::dynexite:
            throw BonusParseException(str::error::unexpectedDynexite);
        case FileType::rwth:
            throw BonusParseException(str::error::unexpectedRwthOnline);
        default:
            // the file might still be invalid, but that will be decided later
            break;
        }
    }

    const auto nrRows = c.data.size();

    // No data aside from header is allowed
    if (nrRows < 2)
        return;

    mRegistrationNr.reserve(static_cast<int>(nrRows) - 1);
    mNrBonusPoints.reserve(static_cast<int>(nrRows) - 1);

    // Get column index
    const auto registrationNrColumn = static_cast<std::size_t>(
        std::ranges::find(c.data[0], "Matrikelnummer") -
        std::ranges::begin(c.data[0]));
    const auto bonusPointColumn = static_cast<std::size_t>(
        std::ranges::find(c.data[0], "Bonuspunkte (Punkte)") -
        std::ranges::begin(c.data[0]));

    // Throw if relevant columns not found
    if (registrationNrColumn >= c.data[0].size())
        throw BonusParseException(str::error::noIdColumn);
    if (bonusPointColumn >= c.data[0].size())
        throw BonusParseException(str::error::noBonusPointColumn);

    // Insert relevant data into lists
    for (std::size_t i = 1; i < nrRows; ++i) {

        // Registration number column sometimes contains an email address
        // instead of a registration number. Todo: Try to match them up via name
        std::optional<int> registrationNr;
        std::optional<Points> bonusPoints;

        // If stoi fails, registrationNr is empty
        try {
            registrationNr = std::stoi(c.data[i][registrationNrColumn]);
        } catch (...) {
        }

        // If stringToPoints fails, bonusPoints is empty
        try {
            bonusPoints = stringToPoints(c.data[i][bonusPointColumn]);
        } catch (...) {
        }

        // Only add ID and points to the lists if both can be successfully
        // parsed
        if (bonusPoints && registrationNr) {
            mRegistrationNr.emplace_back(*registrationNr);
            mNrBonusPoints.emplace_back(*bonusPoints);
        }
        // Write the student to the invalid registration number list
        // instead, to generate a warning later on
        // Ignore the error if the student got no bonus points anyway
        else if (!bonusPoints || (!registrationNr && bonusPoints->p != 0.)) {
            mInvalidRegNrs.emplace_back(c.data[i][registrationNrColumn],
                                        c.data[i][bonusPointColumn]);
        }
    }

    if (mRegistrationNr.size() != mNrBonusPoints.size())
        throw BonusParseException(str::error::bonusArraySizeMismatch);
}
