#include "combined_grade_list.hpp"

#include "user_grade_config.hpp"

#include <algorithm>
#include <cassert>
#include <functional>
#include <ranges>
#include <utility>

[[nodiscard]] auto setWeights(const GradeBoundary closedMax,
                              const GradeBoundary openMax) noexcept {
    const auto combinedMax = closedMax + openMax;
    if (combinedMax.mMultiple <= 0) {
        // Selected exam has neither open nor closed points, so weights don't
        // matter.
        // To avoid division by 0, just set an arbitrary weight.
        return std::make_pair(0.5, 0.5);
    } else {
        return std::make_pair(closedMax / combinedMax, openMax / combinedMax);
    }
}

void CombinedGradeList::recalculate() {

    const auto &closedMaxRef =
        std::get<const ChangedMaxPoints<GradingType::Closed> &>(
            this->mParentRefs);
    const auto &openMaxRef =
        std::get<const ChangedMaxPoints<GradingType::Open> &>(
            this->mParentRefs);

    const auto &closedGradesRef =
        std::get<const GradeList<GradingType::Closed> &>(this->mParentRefs);
    const auto &openGradesRef =
        std::get<const GradeList<GradingType::Open> &>(this->mParentRefs);

    const auto &configRef =
        std::get<const UserGradeConfig &>(this->mParentRefs);

    // Since we want to compute the average for each pair of grades, the
    // two lists have to have the same size.
    assert(closedGradesRef.mGrades.size() == openGradesRef.mGrades.size());

    const auto [closedWeight, openWeight] =
        setWeights(closedMaxRef.mChangedMaxPointBoundary,
                   openMaxRef.mChangedMaxPointBoundary);

    // Assert that the weights are sensible and add to 1.
    assert(closedWeight >= 0. && closedWeight <= 1.);
    assert(openWeight >= 0. && openWeight <= 1.);
    assert(qFuzzyCompare(closedWeight + openWeight, 1.));

    // Resize the QList, reset the grades
    mGrades = QList<Grade>(closedGradesRef.mGrades.size());

    // Transform takes a pair of values from the two input lists at the
    // same index and writes the result of getGradeAverage into the same
    // position in the new list.
    using namespace std::placeholders;
    std::ranges::transform(closedGradesRef.mGrades, openGradesRef.mGrades,
                           std::ranges::begin(mGrades),
                           std::bind(getGradeAverage, _1, closedWeight, _2,
                                     openWeight,
                                     configRef.mRequirePassAllParts));
}
