/**
 * \file csv_parser.cpp
 * \brief Implements functions to read and write csv data, including internal
 * helper functions.
 */

#include "csv_parser.hpp"

#include "util/strings.hpp"

#include <algorithm>
#include <array>
#include <cassert>
#include <fstream>
#include <ios>
#include <ranges>
#include <sstream>
#include <stdexcept>
#include <streambuf>
#include <utility>
#include <vector>

/**
 * \brief To be able to read files with different line endings platform
 * independently, we use this helper function instead of std::getline
 *
 * \author https://stackoverflow.com/a/70582873
 */
std::istream &safeGetLine(std::istream &is, std::string &t) {
    t.clear();

    // The characters in the stream are read one-by-one using a std::streambuf.
    // That is faster than reading them one-by-one using the std::istream.
    // Code that uses streambuf this way must be guarded by a sentry object.
    // The sentry object performs various tasks,
    // such as thread synchronization and updating the stream state.

    std::istream::sentry se(is, true);
    std::streambuf *sb = is.rdbuf();

    for (;;) {
        int c = sb->sbumpc();
        switch (c) {
        case '\n':
            return is;
        case '\r':
            if (sb->sgetc() == '\n')
                sb->sbumpc();
            return is;
        case std::streambuf::traits_type::eof():
            is.setstate(std::ios::eofbit);
            if (t.empty())
                is.setstate(std::ios::failbit);
            return is;
        default:
            t += (char)c;
        }
    }
}

/**
 * \brief Splits the given string into a vector of strings using the given
 * delimiter.
 *
 * \param[in] text      String to split; use `std::move()` to avoid a copy.
 * \param[in] delim     Delimiter character.
 * \param[in] nrDelims  A guess at how many segments exist (optional, does not
 * have to be accurate). \param[in] allowEndWithDelim If the last character in
 * `text` is a delimiter, don't add or add an empty last string to the vector.
 *
 * \author Adopted from
 * https://stackoverflow.com/questions/5607589
 *
 * \return A vector of the string segments between the delimiter characters in
 * original order, with the delimiter characters removed.
 */
std::vector<std::string>
split(std::string text, const char delim, const bool allowEndWithDelim,
      const std::vector<std::string>::size_type nrDelims = 0) {
    assert(delim != '\n' && "For platform-independently splitting a string, "
                            "use safeGetLine instead of split!");

    // reserving enough space beforehand is optional, but improves performance
    std::vector<std::string> vec;
    vec.reserve(nrDelims);

    // pass by rvalue to avoid copying the string
    std::stringstream ss(std::move(text));

    // read until delimiter into line, then move the line to the end of the
    // vector
    std::string line;
    while (std::getline(ss, line, delim)) {
        vec.emplace_back(std::move(line));
    }

    // special case: if last character is the delimiter, we add an empty string
    // as last segment
    if (!allowEndWithDelim && ss.str().back() == delim) {
        vec.emplace_back();
    }

    return vec;
}

/**
 * \brief Counts the number of delimiters in each line, and throws an error if
 * the number is inconsistent.
 *
 * \todo Escaped delimiters.
 *
 * \param[in] fileLines One string per line of the original file
 * \param[in] delim     Delimiter character
 *
 * \exception std::runtime_error The number of delimiters per line is
 * inconsistent.
 *
 * \return Returns the consistent number of delimiters per line.
 */
std::vector<std::string>::size_type
getNrDelims(const std::vector<std::string> &fileLines, const char delim) {
    if (fileLines.empty())
        return 0;

    // take number of delimiters in the first line as baseline
    const auto nrDelims = std::ranges::count(fileLines[0], delim);

    if (nrDelims == 0) {
        throw std::runtime_error("Specified delimiter not used");
    }

    // check that the number of delimiters in every other line is the same
    for (const auto &l : fileLines) {
        if (std::ranges::count(l, delim) != nrDelims) {
            // inconsistent number of delimiters (or escaped ones?)
            throw std::runtime_error(
                "Inconsistent number of delimiter characters");
        }
    }

    return static_cast<std::vector<std::string>::size_type>(nrDelims);
}

std::unique_ptr<CsvDump> readCsv(const std::filesystem::path &p) {
    // try to open the file, throw invalid_argument if file couldn't be opened
    std::ifstream infile(p);
    if (!infile.is_open()) {
        throw std::invalid_argument("Failed to open file");
    }

    // throw std::ios_base::failure on future read errors
    infile.exceptions(std::ifstream::badbit);

    return readCsv(infile);
}

std::unique_ptr<CsvDump> readCsv(const std::string &s) {
    std::istringstream iss(s);
    return readCsv(iss);
}

[[nodiscard]] auto guessDelimiter(const std::vector<std::string> &fileLines) {
    // Possible delimiter characters in decreasing likelihood
    constexpr std::array possibleDelimiters{';', ',', '\t'};

    for (char delim : possibleDelimiters) {

        try {
            return std::make_pair(delim, getNrDelims(fileLines, delim));
        } catch (std::runtime_error &) {
        }
    }
    throw std::runtime_error(str::error::csvDelimiter);
}

std::unique_ptr<CsvDump> readCsv(std::istream &in) {
    // store the resulting dump on the heap, as it might get quite big
    auto dump = std::make_unique<CsvDump>();

    // Read the given file line-by-line into a vector of strings, without caring
    // about the delimiter for now.
    // May throw std::ios_base::failure, which is raised to the caller of this
    // function
    std::vector<std::string> fileLines;
    {
        std::string line;
        while (!safeGetLine(in, line).eof()) {
            fileLines.emplace_back(std::move(line));
        }
    }

    // remove empty lines
    std::erase(fileLines, "");

    const auto nrLines = fileLines.size();

    // obtain number of delimiters per line, or throw std::runtime_error if it
    // isn't consistent
    const auto [delim, nrDelims] = guessDelimiter(fileLines);

    // write text without delimiters into the CsvDump
    dump->data.reserve(nrLines);
    for (auto &fileLine : fileLines) {
        dump->data.push_back(
            split(std::move(fileLine), delim, false, nrDelims));
        assert(dump->data.back().size() == static_cast<unsigned>(nrDelims + 1));
    }

    // remove potential surrounding quotation marks
    for (auto &line : dump->data) {
        for (auto &str : line) {
            if (str.length() >= 2 && str[0] == '\"' && str.back() == '\"') {
                str = str.substr(1, str.length() - 2);
            }
        }
    }

    return dump;
}

void writeCsv(std::ostream &out, const CsvDump &c, const char delim) {

    for (const auto &l : c.data) {

        // Delimiter between all items, but not after last item
        if (!l.empty()) {
            out << l[0];
            for (const auto &v : l | std::views::drop(1)) {
                out << delim << v;
            }
        }

        // Newline after all items, including last
        out << '\n';
    }
}

std::string writeCsv(const CsvDump &c, const char delim) {
    std::stringstream ss;
    writeCsv(ss, c, delim);
    return ss.str();
}

void writeCsv(const std::filesystem::path &p, const CsvDump &c,
              const char delim) {

    // sets failbit on error
    std::ofstream outfile;
    outfile.open(p, std::ios::out | std::ios::trunc);

    // throw std::ios_base::failure on future write errors
    outfile.exceptions(std::ifstream::badbit);

    if (!outfile.fail()) {
        writeCsv(outfile, c, delim);
    } else {
        throw std::invalid_argument{"Failed to open CSV file for write"};
    }
}
