#include "input_model.hpp"

#include "file_io/csv_parser.hpp"
#include "util/defaults.hpp"
#include "util/exception.hpp"
#include "util/strings.hpp"

#include <QList>
#include <QMessageBox>
#include <QString>

InputModel::InputModel(const FileType t) : mType(t) {}

FileOpenError InputModel::loadCsv(const std::filesystem::path &p) {

    // We could move this to a separate thread to keep the interface
    // responsive, but it only takes 1ms on my machine, so let's not
    // bother.
    try {
        FileOpenError error;
        if (mType == FileType::d2r) {

            mLoadedData.emplace<std::unique_ptr<ExamSave>>(readYml(p));

        } else {
            const auto dump = readCsv(p);

            switch (mType) {
            default:
                assert(false);
                [[fallthrough]];
            case FileType::dynexite:
                mLoadedData.emplace<std::unique_ptr<DynexiteExam>>(
                    new DynexiteExam(*dump));
                break;
            case FileType::rwth:
                mLoadedData.emplace<std::unique_ptr<StudentData>>(
                    new StudentData(*dump));
                break;
            case FileType::bonus:
                mLoadedData.emplace<std::unique_ptr<BonusPoints>>(
                    new BonusPoints(*dump));
                for (const auto &regPair :
                     std::get<std::unique_ptr<BonusPoints>>(mLoadedData)
                         ->mInvalidRegNrs) {
                    error.mWarnings.append(
                        QString("Failed to add ") +
                        QString::fromStdString(regPair.second) +
                        " bonus points to student with registration number " +
                        QString::fromStdString(regPair.first));
                }
                break;
            }
        }

        return error;
    } catch (const std::exception &e) {
        switch (mType) {
        default:
            assert(false);
            [[fallthrough]];
        case FileType::dynexite:
            return FileOpenError{FileOpenErrorCode::UnknownError,
                                 QString(str::error::defaultDynexite) +
                                     e.what()};
        case FileType::rwth:
            return FileOpenError{FileOpenErrorCode::UnknownError,
                                 QString(str::error::defaultRwthOnline) +
                                     e.what()};
        case FileType::bonus:
            return FileOpenError{FileOpenErrorCode::UnknownError,
                                 QString(str::error::defaultBonus) + e.what()};
        case FileType::d2r:
            return FileOpenError{FileOpenErrorCode::UnknownError,
                                 QString(str::error::defaultD2R) + e.what()};
        }
    }
}

std::variant<DynexiteExam *, StudentData *, BonusPoints *, ExamSave *>
InputModel::read() const {
    std::variant<DynexiteExam *, StudentData *, BonusPoints *, ExamSave *> ret;
    if (std::holds_alternative<std::unique_ptr<DynexiteExam>>(mLoadedData)) {

        ret = std::get<std::unique_ptr<DynexiteExam>>(mLoadedData).get();
    } else if (std::holds_alternative<std::unique_ptr<StudentData>>(
                   mLoadedData)) {
        ret = std::get<std::unique_ptr<StudentData>>(mLoadedData).get();
    } else if (std::holds_alternative<std::unique_ptr<BonusPoints>>(
                   mLoadedData)) {
        ret = std::get<std::unique_ptr<BonusPoints>>(mLoadedData).get();
    } else {
        ret = std::get<std::unique_ptr<ExamSave>>(mLoadedData).get();
    }
    return ret;
}

void warnIfLowStudentIntersection(QWidget *parent,
                                  const InputModel &dynexiteModel,
                                  const InputModel &rwthModel,
                                  const InputModel &bonusModel,
                                  const FileSelectState &bonusState) noexcept {
    auto dynexiteStudents =
        std::get<DynexiteExam *>(dynexiteModel.read())->mIdentifier;
    auto rwthOnlineStudents =
        std::get<StudentData *>(rwthModel.read())->mRegistrationNr;
    auto bonusStudents =
        bonusState == FileSelectState::ok
            ? std::get<BonusPoints *>(bonusModel.read())->mRegistrationNr
            : QList<int>{};

    std::ranges::sort(dynexiteStudents);
    std::ranges::sort(rwthOnlineStudents);
    std::ranges::sort(bonusStudents);

    QList<int> dbIntersection, drIntersection;

    // Find out number of students that are both in the dynexite file and the
    // RWTHonline/bonus file
    std::ranges::set_intersection(dynexiteStudents, bonusStudents,
                                  std::back_inserter(dbIntersection));
    std::ranges::set_intersection(dynexiteStudents, rwthOnlineStudents,
                                  std::back_inserter(drIntersection));

    QString warning = str::error::lowStudentIntersection;
    QList<QString> singleWarnings;

    // Display warning if few students in dynexite are also in RWTHonline
    if (static_cast<double>(drIntersection.size()) <
        defaults::studentIntersectWarningThreshold *
            static_cast<double>(dynexiteStudents.size())) {
        const auto percentage = static_cast<double>(drIntersection.size()) /
                                static_cast<double>(dynexiteStudents.size()) *
                                100.;

        if (percentage > 0.) {
            singleWarnings.append(
                "\n\nOnly " + QString::number(percentage, 'f', 2) + "% (" +
                QString::number(drIntersection.size()) + " of " +
                QString::number(dynexiteStudents.size()) +
                ") of students in the dynexite file can be found in the "
                "RWTHonline file.");
        } else {
            // Special string for no intersection
            singleWarnings.append(
                "\n\nNone of the students in the dynexite file can be found in "
                "the RWTHonline file.");
        }
    }

    // Display warning if few students in dynexite are also in the bonus point
    // file, but ignore this if no students are in the bonus point file since it
    // is optional
    if (!bonusStudents.empty() &&
        static_cast<double>(dbIntersection.size()) <
            defaults::studentIntersectWarningThreshold *
                static_cast<double>(dynexiteStudents.size())) {
        const auto percentage = static_cast<double>(dbIntersection.size()) /
                                static_cast<double>(dynexiteStudents.size()) *
                                100.;

        if (percentage > 0.) {
            singleWarnings.append(
                "\n\nOnly " + QString::number(percentage, 'f', 2) + "% (" +
                QString::number(dbIntersection.size()) + " of " +
                QString::number(dynexiteStudents.size()) +
                ") of students in the dynexite file can be found in the bonus "
                "point file.");
        } else {
            singleWarnings.append(
                "\n\nNone of the students in the dynexite file can be found in "
                "the bonus point file.");
        }
    }

    // Display all accumulated warnings with a standard prefix
    if (!singleWarnings.empty()) {
        for (const auto &s : singleWarnings) {
            warning += s;
        }
        QMessageBox::warning(parent, "Reading file warnings", warning);
    }
}
