#pragma once

#include "data/bonus_points.hpp"
#include "data/nodes/dynexite_exam.hpp"
#include "data/nodes/student_data.hpp"
#include "file_io/yml_parser.hpp"
#include "util/enums.hpp"

#include <QList>
#include <QWidget>

#include <filesystem>
#include <memory>
#include <utility>
#include <variant>

enum class FileOpenErrorCode {
    Success,
    OsError,
    CsvParseError,
    YmlParseError,
    UnknownError
};

struct FileOpenError {
    FileOpenErrorCode mEc = FileOpenErrorCode::Success;
    QString mDetails{};

    QList<QString> mWarnings{};
};

class InputModel {
public:
    /**
     * \brief Construct a model that controls file loading for the given file
     * type.
     */
    explicit InputModel(FileType t);

    /**
     * \brief Validates that the user selected a correct file and stores the
     * content in the appropriate data structure in `mLoadedData`.
     *
     * \param p Path to the file
     *
     * \return Success
     */
    FileOpenError loadCsv(const std::filesystem::path &p);

    /**
     * \brief Which file type this model stores.
     */
    const FileType mType;

    /**
     * \brief Moves ownership of the stored data to the caller, is empty
     * afterwards.
     * \return The stored data.
     */
    [[nodiscard]] auto get() { return std::move(mLoadedData); }
    [[nodiscard]] std::variant<DynexiteExam *, StudentData *, BonusPoints *,
                               ExamSave *>
    read() const;

private:
    std::variant<std::unique_ptr<DynexiteExam>, std::unique_ptr<StudentData>,
                 std::unique_ptr<BonusPoints>, std::unique_ptr<ExamSave>>
        mLoadedData{};
};

// Display a warning pop-up if too low an amount of students in the dynexite
// file are also in the RWTHonline file todo: noexcept?
void warnIfLowStudentIntersection(QWidget *parent,
                                  const InputModel &dynexiteModel,
                                  const InputModel &rwthModel,
                                  const InputModel &bonusModel,
                                  const FileSelectState &bonusState) noexcept;
