#include "export_rwthonline.hpp"

#include "data/csv_dump.hpp"
#include "data/data_manager.hpp"
#include "file_io/csv_parser.hpp"
#include "util/points.hpp"

#include <QFileDialog>

#include <algorithm>
#include <cctype>
#include <cstddef>
#include <cstdio>
#include <filesystem>
#include <ranges>
#include <string>

// case agnostic string comparison
bool iequals(const std::string &a, const std::string &b) {
    return std::equal(
        a.begin(), a.end(), b.begin(), b.end(),
        [](char c1, char c2) { return std::tolower(c1) == std::tolower(c2); });
}

std::size_t getColumn(const CsvLine &headerLine,
                      const std::string &headerText) {
    // RWTHonline headers are sometimes surrounded by quotation marks
    for (std::size_t i = 0; i < headerLine.size(); ++i) {
        if (iequals(headerLine[i], headerText) ||
            iequals(headerLine[i], "\"" + headerText + '\"'))
            return i;
    }
    return headerLine.size();
}

[[nodiscard]] std::string getRemark(const GradingType type, const int idx) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return "";

    // std::format would be way nicer, but it isn't widely supported yet
    char buf[16] = {};
    std::snprintf(buf, 16, "%.2f", cur->mExam.mPoints[type][idx].p);
    const auto grade = cur->getGradeList(type)[idx];
    return std::string(toString(type)) + " questions: " + buf +
           " pts (grade: " + toString(grade, ',') + ")";
}

void exportKnownStudents(QWidget *parent) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent, "Save RWTHonline data including grades", "",
        "Comma-separated values (CSV) file (*.csv);;All files (*.*)");

    if (!exportLocationString.isEmpty()) {

        auto exportCsvDump = cur->mStudents.mRwthOnline;
        auto &exportCsv = exportCsvDump.data;

        if (exportCsv.empty())
            return;

        // Insert student data into csv
        {
            const auto gradeColumn = getColumn(exportCsv[0], "grade");
            const auto remarkColumn = getColumn(exportCsv[0], "remark");
            if (gradeColumn >= exportCsv[0].size() ||
                remarkColumn >= exportCsv[0].size())
                return;

            // Clear previous cell content
            for (auto &l : exportCsv | std::views::drop(1)) {

                l[gradeColumn] = "";
                l[remarkColumn] = "";
            }

            // Insert new content where we have data
            for (std::size_t i = 0; auto &l : exportCsv | std::views::drop(1)) {

                using enum GradingType;

                // Search for same student in dynexite data
                const auto it = std::ranges::find(
                    cur->mExam.mIdentifier,
                    cur->mStudents.mRegistrationNr[static_cast<qsizetype>(i)]);

                // Convert to index, or skip if not found
                if (it == std::ranges::end(cur->mExam.mIdentifier)) {
                    ++i;
                    continue;
                }
                const auto idx =
                    it - std::ranges::begin(cur->mExam.mIdentifier);

                // Save result in new csv
                l[gradeColumn] = toString(
                    cur->getGradeList(Combined)[static_cast<int>(idx)], ',');
                // Only print closed/open points if there were any to earn
                if (cur->mExam.mPointsMax[Closed] > Points()) {
                    l[remarkColumn] = getRemark(Closed, static_cast<int>(idx));
                }
                if (cur->mExam.mPointsMax[Closed] > Points() &&
                    cur->mExam.mPointsMax[Open] > Points()) {
                    l[remarkColumn] += " - ";
                }
                if (cur->mExam.mPointsMax[Open] > Points()) {
                    l[remarkColumn] += getRemark(Open, static_cast<int>(idx));
                }

                ++i;
            }

            // Mark students without data as missing
            for (auto &l : exportCsv | std::views::drop(1)) {

                if (l[gradeColumn].empty()) {
                    l[gradeColumn] = "X";
                }
            }
        }

        writeCsv(std::filesystem::path(exportLocationString.toStdWString()),
                 exportCsvDump);
    }
}

void exportUnknownStudents(QWidget *parent) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent,
        "Save RWTHonline data including grades of students not in the given "
        "RWTHonline file",
        "", "Comma-separated values (CSV) file (*.csv);;All files (*.*)");

    if (!exportLocationString.isEmpty()) {
        CsvDump exportDump;
        auto &exportCsv = exportDump.data;

        // Header line
        exportCsv = {{"REGISTRATION_NUMBER", "GRADE", "REMARK"}};

        // Insert students not in the input RWTHonline file
        for (std::size_t i = 0; auto &id : cur->mExam.mIdentifier) {

            using enum GradingType;

            // Search for same student in RWTHonline
            const auto it =
                std::ranges::find(cur->mStudents.mRegistrationNr, id);

            // Skip if found
            if (it != std::ranges::end(cur->mStudents.mRegistrationNr)) {
                ++i;
                continue;
            }

            // Save result in new line
            exportCsv.emplace_back(3);
            exportCsv.back()[0] = std::to_string(id);
            exportCsv.back()[1] =
                toString(cur->getGradeList(Combined)[static_cast<int>(i)], ',');
            // Only print closed/open points if there were any to earn
            if (cur->mExam.mPointsMax[Closed] > Points()) {
                exportCsv.back()[2] = getRemark(Closed, static_cast<int>(i));
            }
            if (cur->mExam.mPointsMax[Closed] > Points() &&
                cur->mExam.mPointsMax[Open] > Points()) {
                exportCsv.back()[2] += " - ";
            }
            if (cur->mExam.mPointsMax[Open] > Points()) {
                exportCsv.back()[2] += getRemark(Open, static_cast<int>(i));
            }

            ++i;
        }

        writeCsv(std::filesystem::path(exportLocationString.toStdWString()),
                 exportDump);
    }
}

int getNrUnknownStudents() noexcept {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return 0;

    int count = 0;

    for (const auto &id : cur->mExam.mIdentifier | std::views::drop(1)) {

        const auto it = std::ranges::find(cur->mStudents.mRegistrationNr, id);

        // Not found?
        if (it == std::ranges::end(cur->mStudents.mRegistrationNr)) {
            ++count;
        }
    }

    return count;
}
