#include "students_tab.hpp"

#include "data/data_manager.hpp"
#include "util/grade.hpp"

#include <QMessageBox>
#include <QResizeEvent>

#include <array>

static constexpr std::array tableHeaders = {"Student",      "Closed points",
                                            "Closed Grade", "Open points",
                                            "Open Grade",   "Combined Grade"};

StudentsTab::StudentsTab(QWidget *parent) : QWidget(parent), mTable(this) {

    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &StudentsTab::pointsChanged);
    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &StudentsTab::gradesChanged);
}

void StudentsTab::resizeEvent(QResizeEvent *event) {
    mTable.setFixedSize(event->size());
}

void StudentsTab::pointsChanged() {
    const auto cur = DataManager::getCurrent();

    // Clear content and headers
    mTable.clear();

    if (!cur) {
        return;
    }

    // Column 0: Student IDs
    // Columns 1-2: Closed points+grades (if hasClosed)
    // Columns 3-4: Open points+grades (if hasOpen)
    // Column 5: Combined grade (if hasClosed && hasOpen)
    const bool hasClosed =
        cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
    const bool hasOpen = cur->mExam.mPointsMax[GradingType::Open] > Points{0.};
    mTable.setColumnCount(1 + 2 * static_cast<int>(hasClosed) +
                          2 * static_cast<int>(hasOpen) +
                          static_cast<int>(hasClosed && hasOpen));

    // Set horizontal headers
    {
        int columnIdx = 0;

        mTable.setHorizontalHeaderItem(columnIdx++,
                                       new QTableWidgetItem(tableHeaders[0]));

        if (hasClosed) {

            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[1]));
            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[2]));
        }
        if (hasOpen) {

            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[3]));
            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[4]));
        }

        if (hasClosed && hasOpen) {
            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[5]));
        }
    }

    // Sorting has to be disabled while inserting items
    mTable.setSortingEnabled(false);

    const auto &ids = cur->mExam.mIdentifier;
    const auto &closedPoints = cur->mExam.mPoints[GradingType::Closed];
    const auto &openPoints = cur->mExam.mPoints[GradingType::Open];

    mTable.setRowCount(static_cast<int>(ids.size()));

    // Set student IDs and point numbers
    for (qsizetype row = 0; row < ids.size(); ++row) {

        int columnIdx = 0;

        // Student ID
        {
            const auto idItem = new QTableWidgetItem;
            idItem->setData(Qt::DisplayRole, ids[row]);
            idItem->setFlags(idItem->flags() ^ Qt::ItemIsEditable);
            mTable.setItem(static_cast<int>(row), columnIdx++, idItem);
        }

        // Closed
        if (hasClosed) {
            // Points
            {
                const auto closedPointsItem = new QTableWidgetItem;
                closedPointsItem->setData(Qt::DisplayRole, closedPoints[row].p);
                closedPointsItem->setFlags(closedPointsItem->flags() ^
                                           Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               closedPointsItem);
            }

            // Grade
            {
                const auto closedGradeItem = new QTableWidgetItem;
                closedGradeItem->setFlags(closedGradeItem->flags() ^
                                          Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               closedGradeItem);
            }
        }

        // Open
        if (hasOpen) {
            // Points
            {
                const auto openPointsItem = new QTableWidgetItem;
                openPointsItem->setData(Qt::DisplayRole, openPoints[row].p);
                openPointsItem->setFlags(openPointsItem->flags() ^
                                         Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               openPointsItem);
            }

            // Grade
            {
                const auto openGradeItem = new QTableWidgetItem;
                openGradeItem->setFlags(openGradeItem->flags() ^
                                        Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               openGradeItem);
            }
        }

        // Combined grade
        if (hasClosed && hasOpen) {
            const auto combinedGradeItem = new QTableWidgetItem;
            combinedGradeItem->setFlags(combinedGradeItem->flags() ^
                                        Qt::ItemIsEditable);
            mTable.setItem(static_cast<int>(row), columnIdx++,
                           combinedGradeItem);
        }
    }
    mTable.setSortingEnabled(true);
}

void StudentsTab::gradesChanged(const EnumStorage<bool> &types) {
    const auto cur = DataManager::getCurrent();

    if (!cur) {
        return;
    }

    mTable.setSortingEnabled(false);

    const bool hasClosed =
        cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
    const bool hasOpen = cur->mExam.mPointsMax[GradingType::Open] > Points{0.};

    const auto nrRows = mTable.rowCount();

    // Data structures for error handling
    QList<QString> invalidIds;
    QList<int> notFoundIds;

    // Find corresponding grade to the student id in the row
    for (int i = 0; i < nrRows; ++i) {
        bool ok;
        int id = mTable.item(i, 0)->text().toInt(&ok);

        if (!ok) {
            invalidIds.push_back(mTable.item(i, 0)->text());
            continue;
        }

        const auto idIdx = std::ranges::find(cur->mExam.mIdentifier, id) -
                           std::ranges::begin(cur->mExam.mIdentifier);

        // Found?
        if (idIdx >= cur->mExam.mIdentifier.size()) {
            notFoundIds.push_back(id);
            continue;
        }

        static constexpr auto closedGradeColumn = 2;
        const int openGradeColumn = hasClosed ? 4 : 2;
        static constexpr auto combinedGradeColumn = 5;

        // Set new grade strings
        // Use strings instead of doubles for a fixed number of decimal places
        if (hasClosed && types[GradingType::Closed]) {
            mTable.item(i, closedGradeColumn)
                ->setText(QString::fromStdString(
                    toString(cur->getGradeList(GradingType::Closed)[idIdx])));
        }

        if (hasOpen && types[GradingType::Open]) {
            mTable.item(i, openGradeColumn)
                ->setText(QString::fromStdString(
                    toString(cur->getGradeList(GradingType::Open)[idIdx])));
        }
        if (hasClosed && hasOpen) {
            mTable.item(i, combinedGradeColumn)
                ->setText(QString::fromStdString(
                    toString(cur->getGradeList(GradingType::Combined)[idIdx])));
        }
    }

    // Display warnings if some occurred
    if (!invalidIds.empty() || !notFoundIds.empty()) {
        QString informativeText = "";
        for (const auto &id : invalidIds) {
            informativeText += id + "\n";
        }
        for (const auto &id : notFoundIds) {
            informativeText += QString::number(id) + "\n";
        }

        QMessageBox msgBox(this);
        msgBox.setTextFormat(Qt::TextFormat::MarkdownText);
        msgBox.setWindowTitle("Students-Tab warning.");
        msgBox.setText("**Failed to look up grades for the following\nstudent "
                       "ids in the students tab:**\n");
        msgBox.setInformativeText(informativeText);
        msgBox.setStandardButtons(QMessageBox::Ok);
        msgBox.exec();
    }

    mTable.setSortingEnabled(true);
}
