#include "task_tab.hpp"

#include "data/data_manager.hpp"
#include "point_histogram.hpp"

#include <QBarCategoryAxis>
#include <QBarSeries>
#include <QBarSet>
#include <QBoxPlotSeries>
#include <QBrush>
#include <QCategoryAxis>
#include <QChart>
#include <QChartView>
#include <QFontMetrics>
#include <QLabel>
#include <QValueAxis>

#include <algorithm>
#include <iomanip>
#include <sstream>

TaskTab::TaskTab(QWidget *parent) : QScrollArea(parent) {

    setWindowTitle("Tasks");
    setWidgetResizable(true);

    const auto contents = new QWidget(this);
    mLayout = new QGridLayout(contents);
    setWidget(contents);

    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &TaskTab::pointsChanged);
}

void TaskTab::handleResize(const QSize windowSize,
                           const FontSize largeFontSize) {
    changeFontSize(largeFontSize);
    correctAxisLabels(windowSize, largeFontSize);
}

void TaskTab::changeFontSize(const FontSize largeFontSize) {

    for (auto &pchart : mPointCharts) {
        // adjust x/y axis font size
        {
            auto xaxes = pchart->axes(Qt::Orientation::Horizontal);
            for (auto &xax : xaxes) {

                auto font = xax->labelsFont();
                setFontSize(font, largeFontSize);
                xax->setLabelsFont(font);
            }
        }
        {
            auto yaxes = pchart->axes(Qt::Orientation::Vertical);
            for (auto &yax : yaxes) {

                auto font = yax->labelsFont();
                setFontSize(font, largeFontSize);
                yax->setLabelsFont(font);
            }
        }
    }

    // resize the invisible axis of the box chart, which exists for alignment
    // of the two charts
    for (auto &bchart : mBoxCharts) {
        {
            auto axes = bchart->axes(Qt::Orientation::Horizontal);
            for (auto &ax : axes) {

                auto font = ax->labelsFont();
                setFontSize(font, largeFontSize);
                ax->setLabelsFont(font);
            }
        }
    }
}

// todo: combine with function from gradeHistogram
static void niceYAxis(QValueAxis *axis, const int maxBarHeight,
                      const int maxNrYLabels) {
    /* TODO: use values proportional to total number of grades */

    // TODO: update pointChartYMax

    if (maxNrYLabels < 2) {
        axis->setTickInterval(maxBarHeight);
        axis->setMax(maxBarHeight);
    } else

        if (maxBarHeight < maxNrYLabels) {

        axis->setTickInterval(1);
        axis->setMax(maxBarHeight);

    } else {
        auto newYAxisMax = 10;
        auto tickDiff = maxNrYLabels < 6 ? 5 : 2;

        qDebug() << "niceYAxis tasktab: maxBarHeight:" << maxBarHeight
                 << ", maxNrYLabels:" << maxNrYLabels;

        if (maxBarHeight > 10) {
            tickDiff = 5;
            while (true) {
                if (tickDiff == 0) {
                    qDebug() << "wat";
                }
                if (maxBarHeight <= tickDiff * std::max(1, maxNrYLabels - 1)) {
                    newYAxisMax = static_cast<int>(
                        std::ceil(static_cast<double>(maxBarHeight) /
                                  tickDiff) *
                        tickDiff);
                    break;
                } else {
                    tickDiff *= 2;
                }
            }
        }

        if (axis->tickInterval() != tickDiff)
            axis->setTickInterval(tickDiff);
        if (axis->max() != newYAxisMax)
            axis->setMax(newYAxisMax);
    }
}

static void niceXAxis(QValueAxis *axis, const int maxNrLabels,
                      const double stepSize) {

    const auto largestValue = axis->max();

    if (maxNrLabels < 2) {
        axis->setTickInterval(largestValue);
        return;
    }

    double interval = stepSize;

    while (largestValue / interval + 1 > maxNrLabels) {
        interval *= 2;
    }

    axis->setTickInterval(interval);
}

void TaskTab::correctAxisLabels(const QSize windowSize,
                                const FontSize largeFontSize) {

    qDebug() << "Cell rect 0,0:" << mLayout->cellRect(0, 0);

    const auto cur = DataManager::getCurrent();
    if (!cur) {
        return;
    }

    int yAxisHeightPx = 200;

    for (int i = 0; i < mPointCharts.size(); ++i) {
        // y axis
#if 0
        const auto maxNrYLabels =
            static_cast<int>(pchart->plotArea().height() / (2 * largeFontSize));
        qDebug() << "Plot area height: " << pchart->plotArea().height();
#else
        const auto maxNrYLabels = static_cast<int>(
            yAxisHeightPx / (2 * largeFontSize.getPixelSize()));
#endif

        // find largest bar
        int largestBar = 0;
        auto seriesList = mPointCharts[i]->series();
        if (seriesList.size() != 1) {
            qDebug() << "Wrong number of series in a task point chart";
            assert(false);
            continue;
        }

        auto series = static_cast<QAbstractBarSeries *>(seriesList[0]);

        auto barSetList = series->barSets();
        if (barSetList.size() != 1) {
            qDebug() << "Wrong number of bar sets in a task point bar set";
            assert(false);
            continue;
        }

        auto barSet = barSetList[0];

        for (int barSetIdx = 0; barSetIdx < barSet->count(); ++barSetIdx) {
            if ((*barSet)[barSetIdx] > largestBar) {
                largestBar = static_cast<int>((*barSet)[barSetIdx]);
            }
        }

        {
            auto yaxes = mPointCharts[i]->axes(Qt::Orientation::Vertical);
            for (auto &yax : yaxes) {

                niceYAxis(static_cast<QValueAxis *>(yax), largestBar,
                          maxNrYLabels);
                if (mBoxCharts.size() > i) {

                    auto boxChartAxisList = mBoxCharts[i]->axes(Qt::Horizontal);
                    if (boxChartAxisList.size() != 1) {
                        qDebug() << "Wrong size of axis list of box chart:"
                                 << boxChartAxisList.size();
                        assert(false);
                    } else {
                        auto boxChartAxis =
                            static_cast<QCategoryAxis *>(boxChartAxisList[0]);
                        auto labels = boxChartAxis->categoriesLabels();
                        for (auto &l : labels) {
                            boxChartAxis->remove(l);
                        }
                        boxChartAxis->append(
                            QString::number(static_cast<int>(
                                static_cast<QValueAxis *>(yax)->max())),
                            0);
                    }
                    // axisY->append(QString::number(pointChartYMax), 0);
                } else {
                    qDebug()
                        << "No matching box chart found at index" << i << "!";
                    assert(false);
                }
            }
        }

        // x axis
        {
            auto xaxes = mPointCharts[i]->axes(Qt::Orientation::Horizontal);
            for (auto &xax : xaxes) {

                // get number of characters of maximum label
                auto xValueAx = static_cast<QValueAxis *>(xax);
                const auto maxLabel = QString::number(
                    xValueAx->max(), 'f',
                    nrDecimalPlaces(
                        cur->mUserPointConfig.mGradeBoundaryStepSize));
                qDebug() << "Max label is" << maxLabel;

                auto labelFont = xValueAx->labelsFont();
                QFontMetrics fm(labelFont);
                const auto maxLabelWidth = fm.horizontalAdvance(maxLabel);

                // get width of x-ax
                const auto xAxWidth = static_cast<int>(
                    (windowSize.width() - mLayout->cellRect(0, 0).width()) *
                    0.8);

                const auto maxNrLabels =
                    static_cast<int>(xAxWidth / maxLabelWidth / 1.2);
                qDebug() << "Max nr labels:" << maxNrLabels;

                niceXAxis(
                    xValueAx, maxNrLabels,
                    as<double>(cur->mUserPointConfig.mGradeBoundaryStepSize));
            }
        }
    }

    for (auto &pchart : mPointCharts) {

        // y axis
#if 0
        const auto maxNrYLabels =
            static_cast<int>(pchart->plotArea().height() / (2 * largeFontSize));
        qDebug() << "Plot area height: " << pchart->plotArea().height();
#else
        const auto maxNrYLabels = static_cast<int>(
            yAxisHeightPx / (2 * largeFontSize.getPixelSize()));
#endif

        // find largest bar
        int largestBar = 0;
        auto seriesList = pchart->series();
        if (seriesList.size() != 1) {
            qDebug() << "Wrong number of series in a task point chart";
            assert(false);
            continue;
        }

        auto series = static_cast<QAbstractBarSeries *>(seriesList[0]);

        auto barSetList = series->barSets();
        if (barSetList.size() != 1) {
            qDebug() << "Wrong number of bar sets in a task point bar set";
            assert(false);
            continue;
        }

        auto barSet = barSetList[0];

        for (int i = 0; i < barSet->count(); ++i) {
            if ((*barSet)[i] > largestBar) {
                largestBar = static_cast<int>((*barSet)[i]);
            }
        }

        {
            auto yaxes = pchart->axes(Qt::Orientation::Vertical);
            for (auto &yax : yaxes) {

                niceYAxis(static_cast<QValueAxis *>(yax), largestBar,
                          maxNrYLabels);
            }
        }

        // x axis
        {
            auto xaxes = pchart->axes(Qt::Orientation::Horizontal);
            for (auto &xax : xaxes) {

                // get number of characters of maximum label
                auto xValueAx = static_cast<QValueAxis *>(xax);
                const auto maxLabel = QString::number(
                    xValueAx->max(), 'f',
                    nrDecimalPlaces(
                        cur->mUserPointConfig.mGradeBoundaryStepSize));
                qDebug() << "Max label is" << maxLabel;

                auto labelFont = xValueAx->labelsFont();
                QFontMetrics fm(labelFont);
                const auto maxLabelWidth = fm.horizontalAdvance(maxLabel);

                // get width of x-ax
                const auto xAxWidth =
                    (windowSize.width() - mLayout->cellRect(0, 0).width()) *
                    0.8;

                const auto maxNrLabels =
                    static_cast<int>(xAxWidth / maxLabelWidth / 1.2);
                qDebug() << "Max nr labels:" << maxNrLabels;

                niceXAxis(
                    xValueAx, maxNrLabels,
                    as<double>(cur->mUserPointConfig.mGradeBoundaryStepSize));
            }
        }
    }
}

void TaskTab::pointsChanged() {

    // Delete previous charts
    {
        delete mLayout;
        // The following deletes all child widgets, i.e. the charts and
        // chartViews
        delete takeWidget();
        mPointCharts.clear();
        mBoxCharts.clear();

        const auto contents = new QWidget(this);
        mLayout = new QGridLayout(contents);
        setWidget(contents);
    }

    const auto cur = DataManager::getCurrent();
    if (!cur) {
        return;
    }

    const auto &taskEarnedPoints =
        cur->mExam.mTaskEarnedPoints[GradingType::Combined];
    const auto &taskMaxPoints = cur->mExam.mTaskMaxPoints;
    const auto &taskNames = cur->mExam.mTaskNames;

    for (int i = 0; i < taskEarnedPoints.size(); ++i) {

        // Get all points earned in the task
        auto sortedPoints = taskEarnedPoints[i];
        std::ranges::sort(sortedPoints);

        const auto nrStudents = taskEarnedPoints[i].size();

        // Special case bonus points: hide if no students has bonus points
        if (i == taskEarnedPoints.size() - 1 &&
            taskNames[i] == "Bonus points" && nrStudents == 0) {
            continue;
        }

        // Task  name
        {
            const auto nameLabel = new QLabel(taskNames[i]);
            nameLabel->setWordWrap(true);
            mLayout->addWidget(nameLabel, i * 3, 0, 2, 1);
        }

        if (nrStudents == 0) {

            mLayout->addWidget(
                new QLabel("No student participated in this task"), i * 3, 1, 2,
                1);
        } else {

            // Value of the highest label on the y axis of the point chart.
            // Used to determine how much space to leave to the left of the
            // box-and-whiskers-plot
            int pointChartYMax = 0;

            const auto stepSize =
                as<qreal>(cur->mUserPointConfig.mGradeBoundaryStepSize);
            const auto maxPoints =
                std::max(sortedPoints.back().p, taskMaxPoints[i].p);
            const auto nrBuckets =
                static_cast<int>(std::ceil(maxPoints / stepSize) + 1);

            // Point bucket plot
            {
                auto pointChart = new QChart;
                mPointCharts.push_back(pointChart);

                // For each possible point value, count how many students
                // got that many points
                QList<qreal> nrStudentsPerPoint;
                nrStudentsPerPoint.resize(nrBuckets);

                for (const auto p : taskEarnedPoints[i]) {
                    ++nrStudentsPerPoint[impl::getBarIdx(stepSize, p,
                                                         nrBuckets)];
                }

                auto *taskSeries = new QBarSeries();
                taskSeries->setBarWidth(1.);
                taskSeries->setLabelsVisible(false);

                auto *set = new QBarSet("");
                set->append(nrStudentsPerPoint);

                taskSeries->append(set);

                // x-axis
                {
#if 0
                    const auto axisX = new QBarCategoryAxis();

                    for (int j = 0; j < nrBuckets; ++j) {
                        std::stringstream category;
                        category << std::fixed
                                 << std::setprecision(nrDecimalPlaces(
                                        cur->mUserPointConfig
                                            .mGradeBoundaryStepSize))
                                 << (j * static_cast<double>(stepSize));
                        axisX->append(QString::fromStdString(category.str()));
                    }
#else
                    const auto axisX = new QValueAxis();
                    axisX->setTickType(QValueAxis::TicksDynamic);
                    axisX->setTickAnchor(0);
                    axisX->setTickInterval(1);
                    axisX->setMin(0 - static_cast<double>(stepSize) / 2.);
                    axisX->setMax(nrBuckets * static_cast<double>(stepSize) -
                                  static_cast<double>(stepSize) / 2.);
                    axisX->setLabelFormat(
                        QString("%.") +
                        QString::number(nrDecimalPlaces(
                            cur->mUserPointConfig.mGradeBoundaryStepSize)) +
                        "f");
#endif

                    //// font
                    //{
                    //    auto font = axisX->labelsFont();
                    //    font.setPointSize(defaults::fontSizeBig);
                    //    axisX->setLabelsFont(font);
                    //}

                    pointChart->addAxis(axisX, Qt::AlignBottom);
                }

                // y-axis
                {
                    const auto largestBucket =
                        *std::ranges::max_element(nrStudentsPerPoint);

                    const auto axisY = new QValueAxis();
                    axisY->setTickType(QValueAxis::TicksDynamic);
                    axisY->setLabelFormat("%d");
                    axisY->setTickAnchor(0);

                    axisY->setMin(0.0);

                    // Find sensible tick marks
                    static constexpr int maxTicks = 8;
                    if (static_cast<int>(largestBucket) <= maxTicks) {

                        axisY->setTickInterval(1.);
                        pointChartYMax = static_cast<int>(largestBucket);
                        axisY->setMax(pointChartYMax);
                    } else {
                        // Find ideal number of ticks
                        int bestTickNr = 0;
                        int bestTickNrMax = std::numeric_limits<int>::max();

                        for (int tickCount = maxTicks; tickCount > maxTicks - 3;
                             --tickCount) {
                            // Find smallest multiple of tickCount larger
                            // than largestBucket
                            const auto potentialMax =
                                static_cast<int>(
                                    std::ceil(largestBucket / tickCount)) *
                                tickCount;

                            if (potentialMax - static_cast<int>(largestBucket) <
                                bestTickNrMax -
                                    static_cast<int>(largestBucket)) {

                                bestTickNr = tickCount;
                                bestTickNrMax = potentialMax;
                            }
                        }

                        // bestTickNrMax is a multiple of bestTickNr, so the
                        // result of this is an integer
                        axisY->setTickInterval(
                            static_cast<qreal>(bestTickNrMax / bestTickNr));

                        pointChartYMax = bestTickNrMax;
                        axisY->setMax(pointChartYMax);
                    }

                    //// font
                    //{
                    //    auto font = axisY->labelsFont();
                    //    font.setPointSize(defaults::fontSizeBig - 6);
                    //    axisY->setLabelsFont(font);
                    //}

                    pointChart->addAxis(axisY, Qt::AlignLeft);
                    pointChart->addSeries(taskSeries);
                    taskSeries->attachAxis(axisY);
                }

                pointChart->setAnimationOptions(QChart::NoAnimation);
                pointChart->legend()->hide();

                const auto chartView = new QChartView(pointChart);
                chartView->setMinimumSize(400, 300);
                chartView->setInteractive(false);

                mLayout->addWidget(chartView, i * 3, 1);

                qDebug() << "PointChart Margins:" << pointChart->margins();
            }

            // BoxPlot
            {
                auto boxChart = new QChart;
                mBoxCharts.push_back(boxChart);

                auto *taskSeries = new QBoxPlotSeries;
                auto *set = new QBoxSet;

                // Make sure the different plot elements don't overlap each
                // other
                const qreal minDistance = stepSize * 0.1;

                const qreal median = sortedPoints[nrStudents / 2].p;
                const qreal lowerQuartile = std::min(
                    sortedPoints[nrStudents / 4].p, median - minDistance);
                const qreal lowerExtreme = std::min(
                    sortedPoints.begin()->p, lowerQuartile - minDistance);
                const qreal upperQuartile = std::max(
                    sortedPoints[3 * nrStudents / 4].p, median + minDistance);
                const qreal upperExtreme = std::max(
                    sortedPoints.back().p, upperQuartile + minDistance);

                set->setValue(QBoxSet::Median, median);
                set->setValue(QBoxSet::LowerQuartile, lowerQuartile);
                set->setValue(QBoxSet::UpperQuartile, upperQuartile);
                set->setValue(QBoxSet::LowerExtreme, lowerExtreme);
                set->setValue(QBoxSet::UpperExtreme, upperExtreme);

                taskSeries->setBrush(QBrush(Qt::GlobalColor::green));
                taskSeries->append(set);

                boxChart->addSeries(taskSeries);

                // Add a y axis to get the same offset at the left as the point
                // histogram, for easy point alignment
                {
                    const auto axisY = new QCategoryAxis();
                    axisY->append(QString::number(pointChartYMax), 0);
                    // Set label color to white so that it is hopefully
                    // invisible
                    axisY->setLabelsColor(QColorConstants::White);

                    // chart is rotated 90 degrees, so this is on the left
                    axisY->setLabelsAngle(270);
                    boxChart->addAxis(axisY, Qt::AlignBottom);

                    taskSeries->attachAxis(axisY);

                    //// font
                    //{
                    //    auto font = axisY->labelsFont();
                    //    font.setPointSize(defaults::fontSizeBig - 6);
                    //    axisY->setLabelsFont(font);
                    //}
                }

                // add x axis to specify min and max
                {
                    const auto axisX = new QValueAxis();
                    // Set label format so that the label is invisible
                    axisX->setLabelFormat(" ");
                    axisX->setRange(0. - stepSize / 2.,
                                    maxPoints + stepSize / 2.);
                    axisX->setTickType(QValueAxis::TicksFixed);
                    axisX->setTickCount(2);

                    // chart is rotated 90 degrees, so this is on the left
                    axisX->setLabelsAngle(270);
                    boxChart->addAxis(axisX, Qt::AlignRight);
                    taskSeries->attachAxis(axisX);
                }

                boxChart->setAnimationOptions(QChart::NoAnimation);
                boxChart->legend()->hide();

                const auto chartView = new QChartView(boxChart);
                chartView->setMinimumSize(400, 150);
                chartView->setInteractive(false);

                mLayout->addWidget(chartView, i * 3 + 1, 1);

                chartView->rotate(90);

                // The QT box plot has a little bit more space above it (because
                // of our rotation, to the right of it) than the bar chart has
                // to the right of it. Therefore, reduce the upper (because of
                // our rotation, right) margin by 6px.
                // Furthermore, reduce the left and right because of
                // our rotation, above and below) margin to have less space
                // between box plot and bar chart.
                boxChart->setMargins(QMargins(5, 30, 0, 20));
            }
        }

        // Separator line
        {
            const auto line = new QFrame;
            line->setFrameShape(QFrame::HLine);
            line->setFrameShadow(QFrame::Sunken);
            line->setMinimumHeight(30);

            mLayout->addWidget(line, i * 3 + 2, 0, 1, 2);
        }
    }

    // Ratio of 1:10 between name and chart width
    mLayout->setColumnStretch(0, 1);
    mLayout->setColumnStretch(1, 10);

    // Reduce space between box plot and bar chart
    mLayout->setVerticalSpacing(0);
}
