#include "data/testable_objects.hpp"

#include "data/graph.hpp"
#include "data/nodes/dynexite_exam.hpp"
#include "data/nodes/grade_list.hpp"
#include "data/nodes/student_data.hpp"
#include "data/nodes/user_grade_boundaries.hpp"
#include "file_io/csv_parser.hpp"
#include "util/heuristics.hpp"

#include "data/graph.hpp"

#include <gtest/gtest.h>

#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <ranges>
#include <utility>

using std::filesystem::directory_iterator;

// use Value-Parameterized Tests to execute each test for each DAG
class ConfiguredExamTest
    : public testing::TestWithParam<std::pair<std::string, DAG>> {};

// The other tests rely on ConfiguredExams that are constructed outside of a
// test case, therefore we test the constructor separately here again
TEST(ConfiguredExamTest, createConfiguredExam) {
    for (const auto &dynexiteEntry : directory_iterator(
             std::filesystem::path(TESTING_DYNEXITE_DIRECTORY))) {

        SCOPED_TRACE("Reading in file " + dynexiteEntry.path().string());
        // expect only .csv and .gitignore files in the repository
        if (dynexiteEntry.path().extension() == "") {
            continue;
        }
        if (dynexiteEntry.path().extension() != ".csv") {
            ADD_FAILURE() << "Unknown file extension: "
                          << dynexiteEntry.path().extension();
            continue;
        }

        auto dynexiteDump = readCsv(dynexiteEntry);
        if (!dynexiteDump) {
            ADD_FAILURE() << "dynexite readCsv created a NULL-pointer";
            continue;
        }

        // Read corresponding rwthonline-file
        const auto rwthonlinePath =
            std::filesystem::path(TESTING_RWTHONLINE_DIRECTORY) /
            dynexiteEntry.path().filename();
        if (!std::filesystem::exists(rwthonlinePath)) {
            ADD_FAILURE() << "Found no corresponding RWTHOnline-Path for "
                          << dynexiteEntry.path().filename();
            continue;
        }

        auto rwthonlineDump = readCsv(rwthonlinePath);
        if (!rwthonlineDump) {
            ADD_FAILURE() << "rwthOnline readCsv created a NULL-pointer";
            continue;
        }

        try {
            DynexiteExam dynexite(*dynexiteDump);
            StudentData students(*rwthonlineDump);
            UserGradeConfig gc;
            UserPointConfig pc;
            pc.mGradeBoundaryStepSize = guessGradeBoundaryStepSize(dynexite);
            auto g = DAG(dynexite, students, gc, pc);
        } catch (const std::invalid_argument &e) {
            ADD_FAILURE() << "Ctor of DAG threw: " << e.what();
        }
    }
}

INSTANTIATE_TEST_SUITE_P(ConfiguredExamParameterizedTests, ConfiguredExamTest,
                         testing::ValuesIn(exams()));

// Assert that we don't have any unknown files accidentally parsed
// We might want to remove this test in the future
TEST_P(ConfiguredExamTest, NoUnknownTestInputFiles) {
    const auto &name = GetParam().first;
    if (!(name == "split" || name == "closed" || name == "closed_bonus")) {
        FAIL() << "Input file with unknown name " << name;
    }
}

// Consistent Passing Limit: 0.5*maxClosedPoints <= limit <= 0.6*maxClosedPoints
TEST_P(ConfiguredExamTest, PassingLimitsInRange) {

    const auto &graph = GetParam().second;

    const auto stepSize = graph.mUserPointConfig.mGradeBoundaryStepSize;

    EXPECT_GE(toPoints(graph.getPassingLimit(GradingType::Closed), stepSize).p,
              toPoints(graph.mChangedMaxPointsClosed.mOriginalMaxPointBoundary,
                       stepSize)
                      .p *
                  0.5);
    EXPECT_LE(toPoints(graph.getPassingLimit(GradingType::Closed), stepSize).p,
              toPoints(graph.mChangedMaxPointsClosed.mOriginalMaxPointBoundary,
                       stepSize)
                      .p *
                  0.6);
}

// Passing Limit total points = exam total points
TEST_P(ConfiguredExamTest, PassingLimitsTotalPoints) {

    const auto &graph = GetParam().second;

    const auto stepSize = graph.mUserPointConfig.mGradeBoundaryStepSize;
    const auto rounding = graph.mUserPointConfig.mGradeBoundaryRounding;

    const auto maxClosedBoundary = GradeBoundary{
        graph.mExam.mPointsMax[GradingType::Closed], stepSize, rounding};
    const auto maxOpenBoundary = GradeBoundary{
        graph.mExam.mPointsMax[GradingType::Open], stepSize, rounding};

    EXPECT_EQ(graph.mChangedMaxPointsClosed.mOriginalMaxPointBoundary,
              maxClosedBoundary);
    EXPECT_EQ(graph.mChangedMaxPointsOpen.mOriginalMaxPointBoundary,
              maxOpenBoundary);
}

// Consistent GradeBoundaries: Sorted, highest boundary less than max
// points, 4.0-boundary equal to passing limit
TEST_P(ConfiguredExamTest, GradeBoundariesConsistent) {

    const auto &graph = GetParam().second;

    for (const auto gradingType : {GradingType::Closed, GradingType::Open}) {
        EXPECT_TRUE(std::ranges::is_sorted(
            graph.getUserGradeBoundaries(gradingType), std::ranges::greater()));
        EXPECT_EQ(graph.getUserGradeBoundaries(gradingType)[Grade::grade4_0],
                  graph.getPassingLimit(gradingType));
    }
    EXPECT_LE(
        graph.getUserGradeBoundaries(GradingType::Closed)[Grade::grade1_0],
        graph.mChangedMaxPointsClosed.mOriginalMaxPointBoundary);
    EXPECT_LE(graph.getUserGradeBoundaries(GradingType::Open)[Grade::grade1_0],
              graph.mChangedMaxPointsOpen.mOriginalMaxPointBoundary);
}

// No invalid grade (Grade::NR_GRADES) in the grade list
TEST_P(ConfiguredExamTest, NoInvalidGrades) {
    const auto &graph = GetParam().second;

    for (const auto gradingType :
         {GradingType::Closed, GradingType::Open, GradingType::Combined}) {
        EXPECT_TRUE(std::ranges::find(graph.getGradeList(gradingType),
                                      Grade::NR_VALS) ==
                    std::ranges::end(graph.getGradeList(gradingType)));
    }
}

// More points == better grade
TEST_P(ConfiguredExamTest, PointsGiveGrades) {
    const auto &graph = GetParam().second;

    for (const auto gradingType : {GradingType::Closed, GradingType::Open}) {
        for (int i = 0; i < graph.getGradeList(gradingType).size(); ++i) {
            for (int j = 0; j < graph.getGradeList(gradingType).size(); ++j) {
                if (graph.mExam.mPoints[gradingType][i] >
                    graph.mExam.mPoints[gradingType][j]) {
                    EXPECT_LE(graph.getGradeList(gradingType)[i],
                              graph.getGradeList(gradingType)[j]);
                } else {
                    EXPECT_GE(graph.getGradeList(gradingType)[i],
                              graph.getGradeList(gradingType)[j]);
                }
            }
        }
    }
}

// final grade between open and closed grade
TEST_P(ConfiguredExamTest, GradeBetweenGrades) {
    const auto &graph = GetParam().second;

    for (int i = 0; i < graph.getGradeList(GradingType::Combined).size(); ++i) {
        auto dClosed =
            gradeToDouble(graph.getGradeList(GradingType::Closed)[i]);
        auto dOpen = gradeToDouble(graph.getGradeList(GradingType::Open)[i]);
        auto dCombined =
            gradeToDouble(graph.getGradeList(GradingType::Combined)[i]);
        EXPECT_LE(dCombined, std::max(dClosed, dOpen));
        EXPECT_GE(dCombined, std::min(dClosed, dOpen));
    }
}

// Analyze values of the split test more closely
TEST(ConfiguredExamTest, SplitValues) {
    for (const auto &[name, graph] : exams()) {
        if (name == "split") {
            // todo: this data isn't sensitive...right?
            // DynexiteExam
            EXPECT_EQ(graph.mExam.mAttemptIds[0], "ccm0g32rmuqaebb70bg0");

            EXPECT_EQ(graph.mExam.mIdentifier[0], 300001);
            EXPECT_EQ(graph.mExam.mIdentifier[5], 300007);

            EXPECT_EQ(graph.mExam.mTaskMaxPoints.size(), 44);
            EXPECT_EQ(graph.mExam.mTaskMaxPoints[0].p, 10.);
            EXPECT_EQ(graph.mExam.mTaskMaxPoints[4].p, 6.);
            EXPECT_EQ(graph.mExam.mTaskMaxPoints[42].p, 4.);

            EXPECT_EQ(
                graph.mExam.mTaskEarnedPoints[GradingType::Closed][0][3].p, 8.);
            EXPECT_EQ(graph.mExam.mTaskEarnedPoints[GradingType::Open][4][3].p,
                      4.);
            EXPECT_EQ(
                graph.mExam.mTaskEarnedPoints[GradingType::Combined][0][3].p,
                8.);

            EXPECT_EQ(graph.mExam.mPointsMax[GradingType::Closed].p, 89.);
            EXPECT_EQ(graph.mExam.mPointsMax[GradingType::Open].p, 91.);
            EXPECT_EQ(graph.mExam.mPointsMax[GradingType::Combined].p, 180.);

            EXPECT_EQ(graph.mExam.mPoints[GradingType::Closed][0].p, 86.);
            EXPECT_EQ(graph.mExam.mPoints[GradingType::Open][0].p, 59.);
            EXPECT_EQ(graph.mExam.mPoints[GradingType::Combined][0].p, 145.);

            // UserGradeConfig
            EXPECT_EQ(graph.mUserGradeConfig.mRequirePassAllParts,
                      defaults::requirePassAllParts);

            // UserPointConfig
            EXPECT_EQ(graph.mUserPointConfig.mGradeBoundaryStepSize,
                      GradeBoundaryStepSize::Min1_00);
            EXPECT_EQ(graph.mUserPointConfig.mGradeBoundaryRounding,
                      defaults::boundaryRounding);
            EXPECT_EQ(graph.mUserPointConfig.mCountZerothAttemptAsFirstAttempt,
                      defaults::countZerothAttemptAsFirstAttempt);

            // PassingLimits
            EXPECT_EQ(graph.mChangedMaxPointsClosed.mOriginalMaxPointBoundary
                          .mMultiple,
                      89);
            EXPECT_EQ(graph.mChangedMaxPointsClosed.mChangedMaxPointBoundary
                          .mMultiple,
                      89);
            EXPECT_EQ(graph.getPassingLimit(GradingType::Closed).mMultiple,
                      static_cast<int>(
                          89. * 0.6)); // in this particular unit test, the
                                       // absolute passing limit applies
            EXPECT_EQ(
                graph.mChangedMaxPointsOpen.mOriginalMaxPointBoundary.mMultiple,
                91);
            EXPECT_EQ(
                graph.mChangedMaxPointsOpen.mChangedMaxPointBoundary.mMultiple,
                91);
            EXPECT_EQ(graph.getPassingLimit(GradingType::Open).mMultiple,
                      static_cast<int>(
                          91 * defaults::openPointsStartingPassingLimit));

            // todo: manually calculate grade boundaries and check

            return;
        }
    }

    ADD_FAILURE() << "Split test data not found";
}
